/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "common.h"

#include <stdio.h>
#include <time.h>
#include "lock.h"
#include "adm.h"
#include "adm_add.h"
#include "adm_check.h"
#include "udb_check.h"
#include "adm_cgi.h"
#include "set_uid.h"

extern int errno;


/* CHECK_USER - Check the login id, passwords and fullname for a user.
 * If they are OK, return NULL.  Otherwise, return a pointer to an
 * error message string.  This tries to find all possible errors and
 * return a concatination of all error messages, since we are in a batch
 * mode.  The error message string is in malloced memory and should be
 * freed by the caller (assuming you don't just exit immediately).
 */

char *check_user(char *lid, char *pw1, char *pw2, char *fn)
{
char *error_mesg= (char *)malloc(BFSZ);
char *err= error_mesg;
int len, rc= 0;

    /* Check login */
    if (rc= login_bad(lid, err))
    {
	error_mesg= realloc(error_mesg, (len= strlen(error_mesg)) + BFSZ);
	err= error_mesg + len;
    }

    if (current_passwd(lid) != NULL)
    {
	sprintf(err,"Login %s is already in use.\n",lid);
	error_mesg= realloc(error_mesg, (len= strlen(error_mesg)) + BFSZ);
	err= error_mesg + len;
	rc++;
    }

    /* Check fullname */
    if (fullname_bad(fn, err))
    {
	error_mesg= realloc(error_mesg, (len= strlen(error_mesg)) + BFSZ);
	err= error_mesg + len;
	rc++;
    }

    /* Check password */
    if (pw1 != NULL && pw2 != NULL && strcmp(pw1,pw2))
    {
	strcpy(err,"The two passwords given don't match.\n");
	rc++;
    }
    else
	rc+= passwd_bad(pw1, lid, err);
    
    if (rc == 0)
    {
	free(error_mesg);
	return NULL;
    }
    else
	return error_mesg;
}


/* DIE - Display a message and exit.
 */

#if __STDC__ && HAVE_STDVARARGS
void die(const char *fmt,...)
#else
void die(fmt, va_alist)
    const char *fmt;
    va_dcl
#endif
{
#if HAVE_STDVARARGS
va_list ap;
#endif

    VA_START(ap,fmt);
    set_cfadm();
    fputs("<TITLE>Account Creation Error</TITLE>\n",stdout);
    fputs("<H1>Account Creation Error</H1>\n",stdout);
    fputs("An internal error caused account creation to fail:\n", stdout);
    fputs("<PRE>\n",stdout);
    VA_PRINTF(stdout, fmt, ap);
    fputs("</PRE>\n",stdout);
    VA_END(ap);
    exit(1);
}


/* LOG - Log the account creation.
 */

void log(char *lid, char *fn, char *dir)
{
char *var, *val;
FILE *fp;
time_t clock= time(0L);

	if ((fp= fopen(ACCT_LOG_FILE,"a")) == NULL)
		return;
	lock_exclusive(fileno(fp),ACCT_LOG_FILE);

	fprintf(fp,"-- %15.15s makeacct: %s (%s)\n dir=%s rhost=",
		ctime(&clock)+4,lid,fn,dir);

        if ((val= getenv("REMOTE_HOST")) != NULL)
                fprintf(fp, "%s", val);
        if ((val= getenv("REMOTE_ADDR")) != NULL)
                fprintf(fp, " (%s)", val);
	fputc('\n',fp);

	if (!firstvar(&var, &val))
		return;

	do {
		/* Skip the important stuff */
		if (!strcmp(var,"login") ||
		    !strcmp(var,"fullname") ||
		    !strncmp(var,"passwd",6))
			continue;

		fprintf(fp," %s=%s\n", var, val);

	} while (nextvar(&var,&val));

	fclose(fp);
	unlock(ACCT_LOG_FILE);
}


/* SAVEBIO - Write out bio information -- namely all other fields in query
 */

void savebio(char *dir)
{
char *var, *val;
char fname[BFSZ];
FILE *fp;

	if (!firstvar(&var, &val))
		return;

	sprintf(fname,"%s/.backtalk", dir);
	if ((fp= fopen(fname,"w")) == NULL)
		die("could not create .backtalk file %s (%s)",
			fname, strerror(errno));

	/* Save registration time */
	fprintf(fp,"regdate=%ld\n",time(0L));

	/* Save all other fields in query */
	do {
		/* Skip the important stuff */
		if (!strcmp(var,"login") ||
		    !strcmp(var,"fullname") ||
		    !strncmp(var,"passwd",6))
			continue;

		/* Write the rest into the tag file */
		fprintf(fp,"%s=(", var);
		while (*val != '\0')
		{
		    if ( *val == '\n')
			fputs("\\n",fp);
		    else if (*val == '\\')
			fputs("\\\\",fp);
		    else if (*val == '(')
			fputs("\\(",fp);
		    else if (*val == ')')
			fputs("\\)",fp);
		    else if (*val != '\r')
			fputc(*val,fp);
		    val++;
		}
		fputs(")\n",fp);

	} while (nextvar(&var,&val));

	fclose(fp);
}


/* TYPEFILE - send the contents of a file to stdout.  Return true if it does
 * not exist or is empty.
 */

int typefile(char *filename)
{
FILE *fp;
int ch;
int n= 0;

	if ((fp= fopen(filename,"r")) == NULL)
		return 1;
	
	while ((ch= getc(fp)) != EOF)
	{
		n++;
		putchar(ch);
	}
	
	fclose(fp);
	return (n == 0);
}


/* MAIN */

main(int argc, char **argv)
{
char *lid, *fn, *pw1, *pw2;
char *err, *show;
char *query;
char bf[BFSZ+1], dir[BFSZ];
FILE *rfp;

    umask(022);

    /* Print CGI header */
    fputs("Content-type: text/html\n\n",stdout);

#if ONLY_ADM_CREATE
    if (getuid() != SERVER_UID)
    {
	fputs("<TITLE>Account Application Rejected</TITLE>\n",stdout);
	fputs("<H1>Account Application Rejected</H1>\n",stdout);
	fputs("Program not run by http daemon\n", stdout);
	exit(1);
    }
    if ((lid= getenv("REMOTE_USER")) == NULL ||
        (!inlist(lid,ADM_NAMES,",") &&
        next_uid(0) > 1))
    {
	fputs("<TITLE>Account Application Rejected</TITLE>\n",stdout);
	fputs("<H1>Account Application Rejected</H1>\n",stdout);
	fputs("Only the conference adminstrators can create accounts\n", stdout);
	exit(1);
    }
#endif

    /* Get the query */
#ifdef INSECURE
    if (argc == 2 && !strcmp(argv[1],"-r"))
    {
	if ((rfp= fopen("/tmp/wactr","r")) != NULL)
	{
	    fgets(bf, BFSZ, rfp);
	    if ((query= strchr(bf,'\n')) != NULL) *query= '\0';
	    query= bf;
	    fclose(rfp);
	}
    }
    else
    {
#endif
	query= get_query();
#ifdef INSECURE
	if ((rfp= fopen("/tmp/wactr","w")) != NULL)
	{
	    fputs(query, rfp);
	    fputc('\n',rfp);
	    fclose(rfp);
	}
    }
#endif

    /* Get the user information */
    load_query(query);
    lid= getvar("login");
    fn= getvar("fullname");
    pw1= getvar("passwd1");
    pw2= getvar("passwd2");
    show= getvar("show");

    /* Check user information for sanity */
    if ((err= check_user(lid, pw1, pw2, fn)) != NULL)
    {
	fputs("<TITLE>Account Application Rejected</TITLE>\n",stdout);
	fputs("<H1>Account Application Rejected</H1>\n",stdout);
	fputs("Your account could not be created because:\n", stdout);
	fputs("<PRE>\n",stdout);
	fputs(err,stdout);
	fputs("</PRE>\n",stdout);
	fputs("Please correct your application and submit it again.\n", stdout);
	exit(1);
    }

    /* Create conference account */
    adduser(lid, pw1, fn, dir);		/* this computes dir */
    savebio(dir);

    /* Log the request */
    log(lid,fn,dir);

    /* Print success screen */
    if (show == NULL || typefile(show))
	fputs("Account Created.\n",stderr);

    exit(0);
}
