/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* ROUTINES TO READ ITEM FILES
 *
 * Mar 15, 1993 - Jan Wolter:  Original bbsread version
 * Jun  4, 1993 - Jan Wolter:  Skip scribbled text
 * Dec  2, 1995 - Jan Wolter:  Ansification
 * Jan 10, 1996 - Jan Wolter:  Backtalk version.
 */

#include "backtalk.h"

#include <time.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#include <ctype.h>

#include "sysdict.h"
#include "format.h"
#include "dict.h"
#include "str.h"
#include "ps_item.h"
#include "ps_part.h"
#include "ps_sum.h"
#include "lock.h"

int curr_item= -1;	/* Which item we currently have open (-1 if none) */
FILE *item_fp;		/* File descriptor for current item file */
int reading_resp;	/* number of the response we are reading */
int read_head;		/* Have we read the header for this response yet? */
int resp_type;		/* Type of current resp.  0=plain, 1=html */
char item_path[BFSZ];	/* Full pathname of item file */


/* OPENITEM - Make sure the item named by the "item" system variable in the
 * directory named by the "confdir" system variable is open and ready to
 * read from.  If the excl flag is true, and exclusive lock is set and the
 * file is opened read/write, otherwise a shared lock is set and the item
 * is only opened to read.  Frozen items cannot be opened exclusively.
 * Returns non-zero if the item does not exist.
 */

int openitem(int excl)
{

    if (curr_item != -1 && idict(VAR_ITEM) == curr_item)
	return 0;

    /* Close some other item */
    if (curr_item != -1)
    {
	fclose(item_fp);
	unlock(item_path);
    }

    curr_item= idict(VAR_ITEM);
    sprintf(item_path,"%s/_%d",sdict(VAR_CONFDIR),curr_item);
    if ((item_fp= fopen(item_path,excl?"r+":"r")) == NULL)
    {
	/* If the open fails, there is a tiny chance we hit it in the
	 * middle of a retitle operation.  Wait a bit and try again. */
	sleep(2);
	if ((item_fp= fopen(item_path,excl?"r+":"r")) == NULL)
	{
	    curr_item= -1;
	    return 1;
	}
    }
    lock(excl,fileno(item_fp),item_path);
    reading_resp= 0;
    read_head= FALSE;

    /* Also open the index file */
    open_item_index(sdict(VAR_CONFDIR), curr_item);

    return 0;
}


/* CLOSEITEM - Make sure no item is open.
 */

void closeitem()
{
    if (curr_item != -1)
    {
	fclose(item_fp);
	unlock(item_path);
	curr_item= -1;
    }

    /* Also close the index file */
    close_item_index();
}


/* FIXIMODES - Fix any incorrectly set item modes, both in the system
 * dictionary and in the sum file.
 */

void fiximodes()
{
struct stat st;

    if (fstat(fileno(item_fp),&st))
	die("Unable to stat item");
    
    /* Fix linked status */
    if (st.st_nlink > 1)
    {
	if (!bdict(VAR_LINKED))
	{
	    put_sum(idict(VAR_ITEM),-1,SF_LINKED,0,0L,0L);
	    seti(VAR_LINKED,1);
	}
    }
    else
    {
	if (bdict(VAR_LINKED))
	{
	    put_sum(idict(VAR_ITEM),-1,0,SF_LINKED,0L,0L);
	    seti(VAR_LINKED,0);
	}
    }

    /* Fix frozen status */
    if (!(st.st_mode & 0200))
    {
	if (!bdict(VAR_FROZEN))
	{
	    put_sum(idict(VAR_ITEM),-1,SF_FROZEN,0,0L,0L);
	    seti(VAR_FROZEN,1);
	}
    }
    else
    {
	if (bdict(VAR_FROZEN))
	{
	    put_sum(idict(VAR_ITEM),-1,0,SF_FROZEN,0L,0L);
	    seti(VAR_FROZEN,0);
	}
    }

    /* Fix retired status */
    if (st.st_mode & 0100)
    {
	if (!bdict(VAR_RETIRED))
	{
	    put_sum(idict(VAR_ITEM),-1,SF_RETIRED,0,0L,0L);
	    seti(VAR_FROZEN,1);
	}
    }
    else
    {
	if (bdict(VAR_RETIRED))
	{
	    put_sum(idict(VAR_ITEM),-1,0,SF_RETIRED,0L,0L);
	    seti(VAR_RETIRED,0);
	}
    }
}


/* ITEMHEAD - Read in the header for an item, saving the various values in
 * various sysdict variables.  Returns non-zero if the item does not exist.
 */

int itemhead()
{
char buf[BFSZ];
char *a;

    /* Open the item, if we haven't already */
    if (openitem(0)) return 1;

    fiximodes();

    /* Position file pointer to read the header */
    if (reading_resp != 0 || read_head)
    {
	rewind(item_fp);
	reading_resp= 0;
	read_head= FALSE;
    }

    /* Read and check magic code number */
    if (fgets(buf,BFSZ,item_fp) == NULL || strcmp(buf,"!<ps02>\n"))
	die("Bad magic number in item file %s",item_path);

    while (fgets(buf,BFSZ,item_fp) != NULL)
    {
	if (buf[0] != ',') continue;
	switch(buf[1])
	{
	case 'H':
	    *firstin(buf+2,"\n")= '\0';
	    sets(VAR_TITLE,buf+2,SM_COPY);
	    break;
	case 'A':
	    *firstin(buf+2,"\n")= '\0';
	    sets(VAR_AUTHORNAME,buf+2,SM_COPY);
	    break;
	case 'U':
	    seti(VAR_AUTHORUID,atoi(buf+2));
	    if ((a= strchr(buf+2,',')) == NULL)
		die("Bad ,U syntax in item file %s",item_path);
	    *firstin(a+1,"\n")= '\0';
	    sets(VAR_AUTHORID,a+1,SM_COPY);
	    break;
	case 'R':
	    /* Handle flags later maybe */
	    /* ,R0001 censored */
	    seti(VAR_HIDDEN,buf[5] == '1');
	    /* ,R0003 scribbled */
	    seti(VAR_ERASED,buf[5] == '3');
	    resp_type= buf[4] - '0';
	    sets(VAR_TEXTTYPE, resp_type ? "text/html" : "text/plain", SM_COPY);
	    break;
	case 'D':
	    sett(VAR_DATE,atoth(buf+2));
	    if ((a= strchr(buf+2,' ')) != NULL)
		sett(VAR_EDITDATE,atoth(a+1));
	    else
		sett(VAR_EDITDATE,0);
#ifdef HIDE_HTML
	    if (resp_type == 0)
#endif
		break;
	    /* may drop through */
	case 'T':
	    read_head= TRUE;
	    return 0;
	case 'P':
            /* Ignore this -- probably a Yapp "response to" line, which
             * we aren't handling yet.
             */
            break;
	case ':':
            /* Ignore this -- probably an HTML response commented out for
             * Yapp compatibility.  We'd only be seeing it here if we
             * aren't compiled with HIDE_HTML.
             */
            break;
	default:
	    die("Unexpected ,%c in item %s header",buf[1],item_path);
	}
    }
    die("Premature end of file in %s header",item_path);
}


/* RESPHEAD - Read in the header for a response, saving the various values in
 * various sysdict variables.  The desired response number is given.
 * Return 1 if there is no such item or response.
 *
 * This should not be used for response 0.  Call itemhead instead.
 *
 * If we have both parent pointers and hidden HTML then this gets just a
 * bit messy.  These two tweaks to Picospan file formats don't mesh very
 * cleanly.
 */

int resphead(int resp)
{
char buf[BFSZ];
char *a;
int parent= 0;
#if defined(PARENT_POINTERS) && defined(HIDE_HTML)
long text_offset= 0L;
#endif

    /* Make sure file is opened */
    if (openitem(0)) return 1;

    /* If we are not already at the right response, jump to it */
    if (resp != reading_resp || read_head)
    {
	if ((reading_resp= seek_resp(item_fp,resp)) < 0)
	{
	    read_head= FALSE;
	    return 1;
	}
    }
    else
    	fseek(item_fp, 2L, 1);	/* Skip ",R" */

    /* Read and Parse the rest of the ,R line */
    if (fgets(buf,BFSZ,item_fp) == NULL)
    	return 1;
    seti(VAR_HIDDEN,buf[3] == '1');
    seti(VAR_ERASED,buf[3] == '3');
    resp_type= buf[2] - '0';
    sets(VAR_TEXTTYPE, resp_type ? "text/html" : "text/plain", SM_COPY);

    /* Parse the rest of the header */
    while (fgets(buf,BFSZ,item_fp) != NULL)
    {
#if defined(PARENT_POINTERS) && defined(HIDE_HTML)
	if (text_offset > 0L && (buf[0] != ',' || buf[1] != 'P'))
	{
	    fseek(item_fp, text_offset, 0);
	    goto done;
	}
#endif
	if (buf[0] != ',') continue;
	switch(buf[1])
	{
	case 'A':
	    *firstin(buf+2,"\n")= '\0';
	    sets(VAR_AUTHORNAME,buf+2,SM_COPY);
	    break;
	case 'U':
	    seti(VAR_AUTHORUID,atoi(buf+2));
	    if ((a= strchr(buf+2,',')) == NULL)
		die("Bad ,U syntax in item %s rhead",item_path);
	    *firstin(a+1,"\n")= '\0';
	    sets(VAR_AUTHORID,a+1,SM_COPY);
	    break;
	case 'D':
	    sett(VAR_DATE,atoth(buf+2));
	    if ((a= strchr(buf+2,' ')) != NULL)
		sett(VAR_EDITDATE,atoth(a+1));
	    else
		sett(VAR_EDITDATE,0);
#ifdef HIDE_HTML
	    if (resp_type != 0)
	    {
#ifdef PARENT_POINTERS
		text_offset= ftell(item_fp);
#else
	        goto done;
#endif /*PARENT_POINTERS*/
	    }
#endif /*HIDE_HTML*/
	    break;
#ifdef PARENT_POINTERS
	case 'P':
	    parent= atoi(buf+2);
#ifdef HIDE_HTML
	    if (resp_type != 0)
	    	goto done;
#endif /*HIDE_HTML*/
	    break;
#endif /*PARENT_POINTERS*/
	case 'T':
	    goto done;
	case ':':
            /* Ignore this -- probably an HTML response commented out for
             * Yapp compatibility.  We'd only be seeing it here if we
             * aren't compiled with HIDE_HTML.
             */
            break;
	default:
	    die("Unexpected ,%c in item %s rhead",buf[1],item_path);
	}
    }
    die("Premature end of file in %s rhead",item_path);

    /* Have reached end of headers */
    done:
    seti(VAR_PARENTRESP, parent);
    read_head= 1;
    return 0;
}


/* RESPTEXT - Read in the text of the next response.  Stores the whole text
 * in a malloced string, and returns a pointer to it.  If <format> is 0, it
 * returns it in plain text form.  If <format> is 1, it returns it as HTML.
 * If <format> is 2, it returns the best raw text.
 */

char *resptext(int form)
{
char buf[BFSZ];
char *b;
char *string= NULL;
int l, len= 0, size= 0;
#ifdef HIDE_HTML
int skip= (form == 0 && resp_type == 1);
#endif

    if (!read_head)
	die("Not ready to read_text");

    while (fgets(buf,BFSZ,item_fp) != NULL)
    {
	b= buf;
	if (b[0] == ',')
	{
	    switch (b[1])
	    {
	    case 'R':	/* Sometimes the ,E is missing for some fool reason */
		fseek(item_fp,-strlen(b),1);
		/* drop through */
	    case 'E':
		goto eot;
	    case ',':
		b++;	/* doubled comma means line starts with one comma */
		break;
	    case ':':   /* ,: just gets dropped off line */
		b+= 2;
		break;
#ifdef HIDE_HTML
	    case 'T':
	    	skip= !skip;
	    	if (resp_type == 1)
		    continue;
#endif
	    	/* drop through */
	    default:
		die("Unexpected ,%c in response text of %s",b[1],item_path);
	    }
	}

#ifdef HIDE_HTML
	if (skip) continue;
#endif

	if (len + (l= strlen(b)) > size)
	{
	    size= len + l + BFSZ;
	    if (string == NULL)
		string= (char *)malloc(size+1);
	    else
		string= (char *)realloc(string,size+1);
	}
	strcpy(string+len,b);
	len+= l;
    }
eot: 

    /* Convert plain to HTML */
    if (form == 1 && resp_type == 0 && len > 0)
    {
    	/* Do an "expand" on the string and surround it by <PRE> and </PRE> */
    	char *tmp= format(string, TRUE, TRUE, 79, FALSE);
    	len= strlen(tmp)+12;
    	string= malloc(len+1);
    	strcpy(string, "<PRE>");
    	strcpy(string+5, tmp);
    	free(tmp);
    	strcpy(string+len-7, "</PRE>\n");
    }

    /* Convert HTML to plain (not needed if we had plain text stored) */
#ifndef HIDE_HTML
    else if (form == 0 && resp_type == 1 && len > 0)
    {
    	string= format_html2text(string);
    	len= strlen(string);
    }
#endif /*HIDE_HTML*/

    /* If we aren't blindfolded, mark the item as seen */
    if (!bdict(VAR_BLINDFOLD))
	update_item(idict(VAR_ITEM), reading_resp, time(0L), FALSE);

    /* Set up for reading the next item */
    reading_resp++;
    read_head= FALSE;

    /* Return the string */
    return ((string == NULL) ? strdup("") : (char *)realloc(string,len+1));
}


/* ATOTH -- This interprets a character string as a hex string and returns an
 * time_t equivalent.
 */

time_t atoth(char *str)
{
time_t n=0;

	for( ; isascii(*str) && isxdigit(*str); str++)
		n= n*16 + *str - (isdigit(*str) ? '0'
					        : (isupper(*str) ? 'A'
								 : 'a') - 10);
	return n;
}


/* RETITLE_ITEM -- Change the title of the item given by the system variable
 * "item" in the conference given by the system variable "confdir" to the
 * given string.  A check is made that you have appropriate access.  Returns
 * 2 if the item does not exist or is frozen, 3 if access is denyed, and 0
 * on success.
 */

int retitle_item(char *newtitle)
{
char buf[BFSZ];
char temp_path[BFSZ];
int temp_fd;
FILE *temp_fp;
size_t temp_sz= 0L;
int ch;
struct stat st;
int need_authorship= 0;

    /* Anonymous users can't retitle */
    if (idict(VAR_UID) < 0 || sdict(VAR_ID)[0] == '\0')
	return 3;

    /* Open item to read/write with an exclusive lock on it */
    if (openitem(1)) return 2;

    if (fstat(fileno(item_fp),&st))
	die("Unable to stat item");
    
    /* Adm always can retitle, fairwitness can if item is not linked */
    if (!bdict(VAR_AMFW) || (!bdict(VAR_AMADM) &&
    	(!bdict(VAR_FW_RETITLE) || st.st_nlink > 1)))
    {
	/* Item author can retitle if option is on */
	if (bdict(VAR_AUTHOR_RETITLE))
	    need_authorship= 1;
	else
	{
	    closeitem();
	    return 3;
        }
    }

    /* Clip title at line break */
    *firstin(newtitle,"\n\r")= '\0';

    /* Create temp file */
    sprintf(temp_path,"%s.temp",item_path);
    temp_fd= open(temp_path, O_RDWR|O_CREAT|O_EXCL, 600);
    if (temp_fd < 0)
    {
    	if (errno == EEXIST)
    	    die("%s already exists -- I won't overwrite it",temp_path);
	else
	    die("cannot create %s",temp_path);
    }
    temp_fp= fdopen(temp_fd, "r+");

    /* Check and copy magic code number */
    if (fgets(buf,BFSZ,item_fp) == NULL || strcmp(buf,"!<ps02>\n"))
    {
    	unlink(temp_path);
	die("Bad magic number in item file %s",item_path);
    }
    fputs(buf, temp_fp);
    temp_sz+= strlen(buf);

    /* Fetch old header line and write new one */
    if (fgets(buf,BFSZ,item_fp) == NULL || buf[0] != ',' || buf[1] != 'H')
    {
    	unlink(temp_path);
	die("Missing title line in %s",item_path);
    }
    fputs(",H", temp_fp);
    fputs(newtitle, temp_fp);
    fputc('\n',temp_fp);
    temp_sz+= strlen(newtitle) + 3;

    /* Copy up through the the authorship line, and check that */
    if (need_authorship)
    {
    	for(;;)
    	{
	    if (fgets(buf,BFSZ,item_fp) == NULL)
	    {
		unlink(temp_path);
		die("Missing ,U author line in %s",item_path);
	    }

	    fputs(buf, temp_fp);
	    temp_sz+= strlen(buf);

	    if (buf[0] == ',' && buf[1] == 'U')
	    {
	    	char *p;
	    	uid_t uid;
	    	uid= atoi(buf+2);
	    	if ((p= strchr(buf+3,',')) == NULL)
	    	{
		    unlock(item_path);
		    unlink(temp_path);
		    die("Bad ,U author line in %s",item_path);
	    	}
	    	p++;
	    	*firstin(p,"\n")= '\0';
	    	if (uid != idict(VAR_UID) || strcmp(p,sdict(VAR_ID)))
	    	{
		    closeitem();
		    unlink(temp_path);
		    return 3;
	    	}
	    	break;
	    }
	}
    }

    /* Copy over rest of file */
    while ((ch= getc(item_fp)) != EOF)
    {
	putc(ch,temp_fp);
	temp_sz++;
    }

    /* Check that the copy wasn't munged by running out of disk or something */
    fflush(temp_fp);
    if (fstat(temp_fd,&st) || st.st_size != temp_sz)
    {
    	unlock(item_path);
    	unlink(temp_path);
    	die("Item file write failed -- out of disk space?");
    }

    /* "cp" the temp file into the old file -- we don't want to do a "mv" here
     * because we'd lose any links to the old file.
     */
    rewind(item_fp);
    rewind(temp_fp);
    while ((ch= getc(temp_fp)) != EOF)
	putc(ch,item_fp);
    ftruncate(fileno(item_fp),temp_sz);

    /* Check that the copy succeeded */
    if (fstat(fileno(item_fp),&st) || st.st_size != temp_sz)
    {
    	unlock(item_path);
    	/* Don't delete the temp file -- it's our only full copy */
    	die("Item copy failed -- out of disk space?  "
    	    "Full item text is in %s", temp_path);
    }

    /* Rebuild the index file - it's probably out of date.
     * It'd be nicer do this while copying the file in the loop above, but for
     * now we'll go ahead and make yet another pass through it.  Sigh.
     */
    rebuild_index(item_fp,sdict(VAR_CONFDIR),curr_item);

    /* Close stuff, remove the temp file, and unlock the item file */
    fclose(temp_fp);
    unlink(temp_path);
    closeitem();

    return 0;
}


/* EDIT_RESP -- Change the content of the response designated by the system
 * variables "resp", "item" and "confdir" to the text given by "text" and
 * "texttype".  The author's alias is changed to "alias".  A check is made
 * that you have appropriate access.  Returns 2 if the item or response does
 * not exist or is frozen, 3 if access is denyed, and 0 on success.
 */

int edit_resp()
{
char buf[BFSZ];
char temp_path[BFSZ];
int temp_fd;
FILE *temp_fp, *log_fp;
size_t temp_sz= 0L;
int ch;
struct stat st;
int need_authorship= 0;
int nresp, rstyle, isplain;
char *p;
uid_t ruid;
char *rid;
time_t rdate, clock;
size_t roffset, rsize;
#ifdef HIDE_HTML
char *plain_text;
#endif

    /* Anonymous users can't edit */
    if (idict(VAR_UID) < 0 || sdict(VAR_ID)[0] == '\0')
	return 3;

    /* Open item to read/write with an exclusive lock on it */
    if (openitem(1))
	return 2;

    if (fstat(fileno(item_fp),&st))
	die("Unable to stat item");
    
    /* Adm always can edit, fairwitness can if item is not linked */
    if (!bdict(VAR_AMFW) || (!bdict(VAR_AMADM) &&
    	(!bdict(VAR_FW_EDIT) || st.st_nlink > 1)))
    {
	/* Item author can edit if option is on */
	if (bdict(VAR_AUTHOR_EDIT))
	    need_authorship= 1;
	else
	{
	    closeitem();
	    return 3;
        }
    }

    /* Create temp file */
    sprintf(temp_path,"%s.temp",item_path);
    temp_fd= open(temp_path, O_RDWR|O_CREAT|O_EXCL, 600);
    if (temp_fd < 0)
    {
	unlock(item_path);
    	if (errno == EEXIST)
    	    die("%s already exists -- I won't overwrite it",temp_path);
	else
	    die("cannot create %s",temp_path);
    }
    temp_fp= fdopen(temp_fd, "r+");

    /* Check and copy magic code number */
    if (fgets(buf,BFSZ,item_fp) == NULL || strcmp(buf,"!<ps02>\n"))
    {
    	unlink(temp_path);
	unlock(item_path);
	die("Bad magic number in item file %s",item_path);
    }
    fputs(buf, temp_fp);
    temp_sz+= strlen(buf);

    /* copy up through header of response being changed */
    nresp= 0;
    while (fgets(buf,BFSZ,item_fp) != NULL)
    {
	/* Check if this is the start of our response */
	if (buf[0] == ',' && buf[1] == 'R' && nresp++ == idict(VAR_RESP))
	{
	    /* Save style */
	    rstyle= buf[4]-'0';
	    /* New response is hidden if old one was */
	    seti(VAR_HIDDEN,buf[5] == '1');
	    /* Editing an erased item unerases it */
	    seti(VAR_ERASED,0);

	    /* Get ,U line */
	    if (fgets(buf,BFSZ,item_fp) == NULL ||
	        buf[0] != ',' || buf[1] != 'U')
	    {
		unlink(temp_path);
		unlock(item_path);
		die("Failed to find ,U line in %s",item_path);
	    }

	    /* Extract user name and uid from line */
	    ruid= atoi(buf+2);
	    if ((p= strchr(buf+3,',')) == NULL)
	    {
		unlock(item_path);
		unlink(temp_path);
		die("Bad ,U author line in %s",item_path);
	    }
	    p++;
	    *firstin(p,"\n")= '\0';

	    /* Check if user name and uid are correct */
	    if (need_authorship &&
	        (ruid != idict(VAR_UID) || strcmp(p,sdict(VAR_ID))))
	    {
		closeitem();
		unlink(temp_path);
		return 3;
	    }
	    rid= strdup(p);

	    /* Get ,A line and ignore it */
	    if (fgets(buf,BFSZ,item_fp) == NULL ||
	        buf[0] != ',' || buf[1] != 'A')
	    {
		unlink(temp_path);
		unlock(item_path);
		die("Failed to find ,A line in %s",item_path);
	    }

	    /* Get ,D line and extract date from it */
	    if (fgets(buf,BFSZ,item_fp) == NULL ||
	        buf[0] != ',' || buf[1] != 'D')
	    {
		unlink(temp_path);
		unlock(item_path);
		die("Failed to find ,D line in %s",item_path);
	    }
	    rdate= atoth(buf+2);

	    clock= time(0L);

#ifdef CENSOR_LOG
	    /* Start writing the logfile entry */
	    log_fp= fopen(CENSOR_LOG,"a");
	    if (log_fp)
	    	fprintf(log_fp, "response %s:%d:%d (%s) edited by (%s) on %s",
			sdict(VAR_CONF), idict(VAR_ITEM), idict(VAR_RESP),
			rid, sdict(VAR_ID), ctimez(&clock,NULL));
#endif

	    /* Skip rest of old response */
	    for (;;)
	    {
		if (fgets(buf,BFSZ,item_fp) == NULL)
		    die("Unterminated response in %s",item_path);
	    	if (buf[0] == ',' &&  (buf[1] == 'E' || buf[1] == 'R'))
		    break;
#ifdef CENSOR_LOG
		if (log_fp) fputs(buf,log_fp);
#endif
	    }
#ifdef CENSOR_LOG
	    if (log_fp) { fputs(",E\n",log_fp); fclose(log_fp); }
#endif

	    /* Prepare new response */
	    isplain= (strcasecmp(sdict(VAR_TEXTTYPE),"text/html") != 0);
	    if (!isplain)
	    {
	    	/* Sanitize HTML */
		sysdict[VAR_TEXT].t.val=
		    (void *)format_html(dup_dict(VAR_TEXT));
		if (sysdict[VAR_TEXT].t.val == NULL)
		    die("HTML cleanup failed");
#ifdef HIDE_HTML
		/* Construct plain text version */
		plain_text= format_html2text(strdup(sysdict[VAR_TEXT].t.val));
		if (plain_text == NULL)
		    die("conversion of HTML to plaintext failed");
#endif
	     }

	    /* Write new response */
#ifdef HIDE_HTML
	    if (!isplain)
	    {
		write_text(temp_fp, ruid, rid, sdict(VAR_ALIAS), 1, plain_text,
		    sdict(VAR_TEXT), rdate, clock, idict(VAR_PARENTRESP),
		    &roffset, &rsize);
		free(plain_text);
	    }
	    else
#endif
		write_text(temp_fp, ruid, rid, sdict(VAR_ALIAS), !isplain,
		    sdict(VAR_TEXT), NULL, rdate, clock, idict(VAR_PARENTRESP),
		    &roffset, &rsize);

	    free(rid);
	    temp_sz+= rsize;
	}
	else
	{
	    /* Lines of other responses and stuff - just copy them */
	    fputs(buf, temp_fp);
	    temp_sz+= strlen(buf);
	}
    }

    /* Check that the copy wasn't munged by running out of disk or something */
    fflush(temp_fp);
    if (fstat(temp_fd,&st) || st.st_size != temp_sz)
    {
    	unlock(item_path);
    	unlink(temp_path);
    	die("Item file write failed -- out of disk space?");
    }

    /* "cp" the temp file into the old file -- we don't want to do a "mv" here
     * because we'd lose any links to the old file.
     */
    rewind(item_fp);
    rewind(temp_fp);
    while ((ch= getc(temp_fp)) != EOF)
	putc(ch,item_fp);
    ftruncate(fileno(item_fp),temp_sz);

    /* Check that the copy succeeded */
    if (fstat(fileno(item_fp),&st) || st.st_size != temp_sz)
    {
    	unlock(item_path);
    	/* Don't delete the temp file -- it's our only full copy */
    	die("Item copy failed -- out of disk space?  "
    	    "Full item text is in %s", temp_path);
    }

    /* Rebuild the index file - it's almost certainly out of date.
     * It'd be nicer do this while copying the file in the loop above, but for
     * now we'll go ahead and make yet another pass through it.  Sigh.
     */
    rebuild_index(item_fp,sdict(VAR_CONFDIR),curr_item);

    /* Close stuff, remove the temp file, and unlock the item file */
    fclose(temp_fp);
    unlink(temp_path);
    closeitem();

    return 0;
}

/* ITEM_POWERS - After opening a new item and setting all the descriptive
 * variables (like "frozen", "linked", "authorid", etc), this routine can
 * be called to set "mayfreeze", "mayretire", "mayretitle" and "maykill" to
 * the appropriate values for the item, and also set "mayhide" and "mayshow"
 * to the values for the item text.
 *
 * The various variables in here have values with the following meanings:
 *   -1    We haven't decided yet if the user has this power
 *    0    The user doesn't have this power
 *    1    The user has this power because he is the item author
 *    2    The user has this power because he is a fairwitness or admin
 */

void item_powers()
{
int freeze= -1, retire= -1, kill= -1;
int retitle= -1, erase= -1, hide= -1, edit= -1;

    /* Due to technical difficulties, impossible to do much with frozen items */
    if (bdict(VAR_FROZEN))
    {
    	retitle= 0;
    	erase= 0;
    	hide= 0;
	edit= 0;
    }

    /* Impossible to hide or unerase responses that are erased */
    if (bdict(VAR_ERASED))
    {
    	erase= 0;
    	hide= 0;
    }

    /* Authors have all sorts of magical powers sometimes*/
    if (idict(VAR_AUTHORUID) == idict(VAR_UID) &&
        !strncmp(sdict(VAR_AUTHORID),sdict(VAR_ID),8))
    {
	if (bdict(VAR_AUTHOR_FREEZE) && freeze == -1) freeze= 1;
	if (bdict(VAR_AUTHOR_RETIRE) && retire == -1) retire= 1;
	if (bdict(VAR_AUTHOR_KILL) && idict(VAR_MAXRESP) == 0 && kill == -1)
		kill= 1;
	if (bdict(VAR_AUTHOR_RETITLE) && retitle == -1) retitle= 1;
	if (bdict(VAR_AUTHOR_ERASE) && erase == -1) erase= 1;
	if (bdict(VAR_AUTHOR_HIDE) && hide == -1) hide= 1;
	if (bdict(VAR_AUTHOR_EDIT) && edit == -1) edit= 1;
    }

    /* Administrators and Fairwitnesses can leap tall buildings */
    if (bdict(VAR_AMFW))
    {
    	if (kill == -1) kill= 2;

    	if (bdict(VAR_AMADM) || !bdict(VAR_LINKED))
    	{
	    if (freeze == -1) freeze= 2;
	    if (retire == -1) retire= 2;
	    if (retitle == -1) retitle= 2;
    	}
    	if ((bdict(VAR_AMADM) || (bdict(VAR_FW_ERASE) && !bdict(VAR_LINKED))) &&
    	    erase == -1) erase= 2;
    	if ((bdict(VAR_AMADM) || (bdict(VAR_FW_HIDE) && !bdict(VAR_LINKED))) &&
    	     hide == -1) hide= 2;
    	if ((bdict(VAR_AMADM) || (bdict(VAR_FW_EDIT) && !bdict(VAR_LINKED))) &&
    	     edit == -1) edit= 2;
    }

    /* Plain ordinary users haven't got much */
    if (freeze == -1) freeze= 0;
    if (retire == -1) retire= 0;
    if (kill == -1) kill= 0;
    if (retitle == -1) retitle= 0;
    if (erase == -1) erase= 0;
    if (hide == -1) hide= 0;
    if (edit == -1) edit= 0;

    /* Go and set the variables */
    seti(VAR_MAYFREEZE, freeze);
    seti(VAR_MAYRETIRE, retire);
    seti(VAR_MAYKILL, kill);
    seti(VAR_MAYRETITLE, retitle);
    seti(VAR_MAYERASE, erase);
    seti(VAR_MAYHIDE, hide);
    seti(VAR_MAYEDIT, edit);
}


/* RESP_POWERS - After opening a new response in an item and setting all the
 * descriptive variables (like "hidden", "erased", "authorid", etc), this
 * routine can be called to set "mayhide", "mayedit" and "mayshow" to the
 * values for that response.
 */

void resp_powers()
{
int erase= -1, hide= -1;    /* -1==dunno, 0==no, 1==yes(author), 2=yes(fw) */
int edit= -1;

    /* Impossible to do anything to frozen items */
    if (bdict(VAR_FROZEN))
    {
    	edit= 0;
	erase= 0;
	hide= 0;
    }

    /* Can't erase or hide erased items (but can edit them) */
    if (bdict(VAR_ERASED))
    {
	erase= 0;
	hide= 0;
    }

    /* Authors sometimes may */
    if (idict(VAR_AUTHORUID) == idict(VAR_UID) &&
        !strncmp(sdict(VAR_AUTHORID),sdict(VAR_ID),8))
    {
	if (bdict(VAR_AUTHOR_ERASE) && erase == -1) erase= 1;
	if (bdict(VAR_AUTHOR_HIDE) && hide == -1) hide= 1;
	if (bdict(VAR_AUTHOR_EDIT) && edit == -1) edit= 1;
    }

    /* Administrators may */
    if (bdict(VAR_AMADM))
    {
	if (hide == -1) hide= 2;
	if (erase == -1) erase= 2;
	if (edit == -1) edit= 2;
    }

    /* Fairwitnesses sometimes may */
    if (bdict(VAR_AMFW))
    {
	if (erase == -1 && bdict(VAR_FW_ERASE) && !bdict(VAR_LINKED)) erase= 2;
	if (hide == -1 && bdict(VAR_FW_HIDE) && !bdict(VAR_LINKED)) hide= 2;
	if (edit == -1 && bdict(VAR_FW_EDIT) && !bdict(VAR_LINKED)) edit= 2;
    }

    /* Strangers may not */
    if (erase == -1) erase= 0;
    if (hide == -1) hide= 0;
    if (edit == -1) edit= 0;

    /* Go and set the variables */
    seti(VAR_MAYERASE, erase);
    seti(VAR_MAYHIDE, hide);
    seti(VAR_MAYEDIT, edit);
}
