/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* ROUTINES TO WRITE ITEM FILES
 * (retitle is in ps_item.c for some reason)
 *
 * A lot of these routines open item files on their own, instead of using
 * the openitem() mechanism.  I've had problems with this causing internal
 * deadlocks, so I made sure they all call closeitem() before opening anything.
 * But I'd be much happier if all items were opened my a single routine.  We
 * should never have more than one open at a time, no matter what people put
 * in the script files.
 */

#include "backtalk.h"

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <time.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#include <sys/stat.h>
#include <errno.h>

#include "date.h"
#include "sysdict.h"
#include "str.h"
#include "format.h"
#include "ps_conflist.h"
#include "ps_post.h"
#include "ps_part.h"
#include "ps_sum.h"
#include "dict.h"
#include "lock.h"


char item_file[BFSZ];

/* WRITE_TEXT_BLOCK - write the part of the item or response text that goes
 * between the ",T" and the ",E" lines.  Really just copy text into file,
 * doubling any initial commas, stripping out carriage returns, and making
 * sure it ends with a newline character.  Return size of the block.
 */

long write_text_block(FILE *fp, char *text,int is_alt)
{
    int nl= 1;
    long size= 0L;

    for (; *text != '\0'; text++)
    {
    	if (*text != '\r')		 /* strip out CR characters */
    	{
#ifdef YAPP_COMPAT
	    /* For Yapp, prefix each line of an HTML response with ,:
	     * Yapp won't tolerate any line outside ,T and ,E having
	     * an upper case letter as its second character, so we insert the
	     * ,: to make sure that doesn't happen.  We throw out the ,:
	     * when reading.
	     */
	    if (is_alt && nl)
	    {
	    	fputc(',',fp);
	    	fputc(':',fp);
	    	size += 2;
	    }
	    else
#endif /* YAPP_COMPAT */
    	    /* double inital commas */
    	    if (nl && *text == ',')
    	    {
    	    	fputc(',',fp);
		size++;
    	    }

    	    /* nl is true if last character written was a newline */
    	    nl= (*text == '\n');

    	    fputc(*text,fp);
	    size++;
    	}
    }

    /* Make sure we have a trailing newline */
    if (!nl)
    {
    	fputc('\n',fp);
    	size++;
    }

    return size;
}


/* WRITE_TEXT - write item or response text to the given file in PicoSpan
 * format.  <fp> is the open file descriptor to write to.  If there is an
 * alternate version of the text, it is passed in as <alt_text>, otherwise
 * this is NULL.  Return the offset of the ",R" in the file, and the total
 * size of the response.
 */

void write_text(FILE *fp, uid_t uid, char *id, char *alias,
		int style, char *text, char *alt_text, time_t postdate,
		time_t editdate, int parent, size_t *offset, size_t *size)
{
char *p;
int nl;
#ifndef FPRINTF_RETURNS_SIZE
char bf[BFSZ];
#endif

    /* Write response header */
    fputc(',',fp);
    *offset= (size_t)ftell(fp)-1;
    fprintf(fp,"R00%c%c\n",
	'0' + style,
	(bdict(VAR_HIDDEN) ? '1' : (bdict(VAR_ERASED) ? '3' : '0')));
    *size= 7;

#ifdef FPRINTF_RETURNS_SIZE
    *size+= fprintf(fp,",U%d,%s\n",uid,id);
    *size+= fprintf(fp,",A%s\n",alias);
#ifndef NO_EDIT_DATE
    if (editdate != 0)
    	*size+= fprintf(fp,",D%lx %lx\n",postdate,editdate);
    else
#endif /*NO_EDIT_DATE*/
    	*size+= fprintf(fp,",D%lx\n",postdate);
#ifdef PARENT_POINTERS
    if (parent > 0)
    	*size+= fprintf(fp,",P%d\n",parent);
#endif /*PARENT_POINTER*/
#else
    sprintf(bf,",U%d,%s\n",uid,id);
    *size+= strlen(bf);
    fputs(bf,fp);
    fprintf(fp,",A%s\n",alias);
    *size+= 3 + strlen(alias);
#ifndef NO_EDIT_DATE
    if (editdate != 0)
	sprintf(bf,",D%lx %lx\n",postdate,editdate);
    else
#endif /*NO_EDIT_DATE*/
	sprintf(bf,",D%lx\n",postdate);
    *size+= strlen(bf);
    fputs(bf,fp);
#ifdef PARENT_POINTERS
    if (parent > 0)
    {
    	sprintf(bf,",P%d\n",parent);
        *size+= strlen(bf);
        fputs(bf,fp);
    }
#endif /*PARENT_POINTER*/
#endif /*FPRINTF_RETURNS_SIZE */

    if (alt_text != NULL)
	*size+= write_text_block(fp,alt_text,1);

    /* Write text */
    fprintf(fp,",T\n");
    *size+= 3;
    *size+= write_text_block(fp,text,0);
    fprintf(fp,",E\n");
    *size+= 3;
    return;
}


/* POST_RESP - Append a response to an item and update the sum file.
 * Item text, user identification, and erased or hidden status are taken
 * from the system dictionary.  Returns new response number or 0 on failure,
 * or -1 if "expect_resp" is set to a positive value that does not match the
 * number the new response would have gotten.
 */

int post_resp()
{
FILE *ifp;
int item,ifd,rn;
int isplain;
time_t filetime, clocktime= time(0L);
struct stat st;
size_t ioffset,size;
#ifdef HIDE_HTML
char *plain_text;
#endif

    /* If text is HTML, sanitize it and possible generate plaintext version */
    isplain= (strcasecmp(sdict(VAR_TEXTTYPE),"text/html") != 0);
    if (!isplain)
    {
    	/* Sanitize HTML */
    	sysdict[VAR_TEXT].t.val= (void *)format_html(dup_dict(VAR_TEXT));
    	if (sysdict[VAR_TEXT].t.val == NULL)
    		die("HTML cleanup failed");
#ifdef HIDE_HTML
	/* Construct plain text version */
	plain_text= format_html2text(strdup(sysdict[VAR_TEXT].t.val));
	if (plain_text == NULL)
		die("conversion of HTML to plaintext failed");
#endif
    }

    /* Make sure no other item is open, to avoid deadlocking against ourself */
    closeitem();

    if (idict(VAR_UID) < 0 || sdict(VAR_ID)[0] == '\0')
	die("Anonymous user attempted to post response");

    item= idict(VAR_ITEM);
    sprintf(item_file,"%s/_%d",sdict(VAR_CONFDIR),item);

    /* Check that item is not frozen */
    if (stat(item_file,&st))
	return 0;
    if (!(st.st_mode & 0200))
	die("Item is frozen");

    /* Open and Lock the item file */
    if ((ifd= open(item_file,O_WRONLY|O_APPEND)) < 0)
	return 0;
    lock_exclusive(ifd,item_file);
    ifp= fdopen(ifd,"a");

    /* If "expect_resp" is postive, but not equal to next response, abort */
    if (idict(VAR_EXPECT_RESP) > 0)
    {
    	if (get_sum(item, &rn, NULL, NULL, NULL))
    	    die("item vanished while responding");
        if (idict(VAR_EXPECT_RESP) < rn)
        {
            fclose(ifp);
	    unlock(item_file);
            return -1;
        }
    }

    /* Write response */
#ifdef HIDE_HTML
    if (!isplain)
    {
	write_text(ifp,idict(VAR_UID),sdict(VAR_ID),sdict(VAR_ALIAS),
		1,plain_text,sdict(VAR_TEXT),clocktime,0,idict(VAR_PARENTRESP),
		&ioffset,&size);
	free(plain_text);
    }
    else
#endif
	write_text(ifp,idict(VAR_UID),sdict(VAR_ID),sdict(VAR_ALIAS),
		!isplain,sdict(VAR_TEXT),NULL,clocktime,0,idict(VAR_PARENTRESP),
		&ioffset,&size);

    fflush(ifp);	/* Get it written out before we do the put_sum */

    /* Get file modification date */
    if (fstat(ifd,&st))
    {
    	fclose(ifp);
    	unlock(item_file);
        die("stat of item failed: %s",strerror(errno));
    }
    filetime= st.st_mtime;

    /* Check that it all got written out */
    if (st.st_size < ioffset+size)
    {
    	/* Chop off any incomplete response */
    	ftruncate(ifd, ioffset);

	fclose(ifp);
	unlock(item_file);
    	die("could not write response - disk full?");
    }

    /* Update Sum file (using time from stat) */
    rn= put_sum(item, -2, 0, 0, 0L, filetime);

    /* Release file */
    fclose(ifp);
    unlock(item_file);

    /* Update index file */
    index_new_resp(sdict(VAR_CONFDIR), item, rn-1, ioffset-2);

    /* Update Participation file (using time from stat) */
    update_item(item, -3, filetime, FALSE);

    return rn-1;
}


/* WRITE_ITEM - Write out a new picospan-format item.  Item text, title,
 * user identification, and erased or hidden status are taken from the system
 * dictionary.  Dies on failure.  Returns offset of comma in ",R".
 */

size_t write_item(FILE *fp, int style, char *text, char *alt_text,
	time_t intime, time_t *extime)
{
struct stat st;
size_t offset,size,hsize;
	
    /* Write item file */
    fprintf(fp,"!<ps02>\n");
    fprintf(fp,",H%s\n",sdict(VAR_TITLE));
    hsize= 11 + strlen(sdict(VAR_TITLE));
    write_text(fp,idict(VAR_UID),sdict(VAR_ID),sdict(VAR_ALIAS),
	       style,text,alt_text,intime,0,0,&offset,&size);
    fflush(fp);

    /* Get file modification date */
    if (fstat(fileno(fp),&st))
    {
    	fclose(fp);
    	unlock(item_file);
        die("stat of item failed: %s",strerror(errno));
    }

    /* Check that it all got written out */
    if (st.st_size < size + hsize)
    {
    	/* Delete any incompletely posted item */
    	unlink(item_file);

    	fclose(fp);
    	unlock(item_file);
    	die("could not write item %s - disk full?",item_file);
    }

    /* Return last modification date of file */
    *extime= st.st_mtime;

    return offset;
}


/* POST_ITEM - Enter a new item and update the sum file.  Item text, title,
 * user identification, and erased or hidden status are taken from the system
 * dictionary.   The number of the item created is saved in the "item"
 * system variable.
 */

void post_item()
{
extern int private_item;
int item,ifd,isplain;
FILE *ifp;
time_t clocktime= time(0L), filetime;
size_t offset;
#ifdef HIDE_HTML
char *plain;
#endif
	
    /* Make sure no other item is open, just in case of deadlock potential */
    closeitem();

    if (idict(VAR_UID) < 0 || sdict(VAR_ID)[0] == '\0')
	die("Anonymous user attempted to post item");

    if (strchr(sdict(VAR_TITLE),'\n') != NULL)
	die("Illegal character in item title");

    /* Response is plaintext if texttype is anything but "text/html" */
    isplain= (strcasecmp(sdict(VAR_TEXTTYPE),"text/html") != 0);

    if (!isplain)
    {
        /* For HTML responses, always run them through format_html */
	sysdict[VAR_TEXT].t.val= (void *)format_html(dup_dict(VAR_TEXT));
	if (sysdict[VAR_TEXT].t.val == NULL)
	    die("html cleanup failed");
#ifdef HIDE_HTML
        /* If we are running text compatible generate a text version */
	if ((plain= format_html2text(strdup(sysdict[VAR_TEXT].t.val))) == NULL)
	    die("html to plain text conversion failed");
#endif
    }

    /* Find and create the next available item file */
    item= last_no() + 1;
    if (item == 1 && !bdict(VAR_AMFW))
	die("Only fairwitness can post first item");

    for (;;)
    {
	/* Try opening the standard item file first */
	sprintf(item_file,"%s/_%d",sdict(VAR_CONFDIR),item);
	ifd= open(item_file, O_WRONLY|O_CREAT|O_EXCL,
		private_item ? 0600 : 0644);
	if (ifd < 0 && errno != EEXIST)
	    die("Cannot create item file %s",item_file);
	if (ifd >= 0)
	{
		/* Was able to create item file */
		break;
	}
	item++;
    }
    lock_exclusive(ifd,item_file);
    ifp= fdopen(ifd,"w");

    /* Write item file */
#ifdef HIDE_HTML
    if (!isplain)
    {
    	offset= write_item(ifp,1,plain,sdict(VAR_TEXT),clocktime,&filetime);
    	free(plain);
    }
    else
#endif
        offset= write_item(ifp,!isplain,sdict(VAR_TEXT),NULL,clocktime,
           &filetime);

    /* Write sumfile */
    (void)put_sum(item, 1, SF_ACTIVE, ~SF_ACTIVE, filetime, filetime);

    /* Release file */
    fclose(ifp);
    unlock(item_file);

    /* Update index file */
    index_new_resp(sdict(VAR_CONFDIR), item, 0, offset);

    /* Update Participation file */
    update_item(item, -1, filetime, FALSE);

    seti(VAR_ITEM,item);
}


/* OPEN_RESP - Open an item, and move to the first character after the ",R"
 * line of the response "resp".  "resp" may be zero, in which case we go to
 * the item header.  Returns NULL if no such item or response.  If write is
 * true, we must have write access.  Also returns the response flags (if
 * rcode is not NULL), the author uid and the author id (if id is not NULL).
 */

FILE *open_resp(int item, int resp, int write, char *rcode, char *id,
    uid_t *uid)
{
char bf[BFSZ+1];
long off;
int  fd;
char *a;
FILE *fp;

    /* Close any other open items, to prevent deadlocks */
    closeitem();

    sprintf(item_file,"%s/_%d",sdict(VAR_CONFDIR),item);
    open_item_index(sdict(VAR_CONFDIR),item);

    /* Open and Lock the file -- frozen items will be opened read-only */
    if ((fd= open(item_file,O_RDWR)) < 0)
    {
        if (write) return(NULL);
        if ((fd= open(item_file,O_RDONLY)) < 0)
		return(NULL);
	lock_shared(fd,item_file);
	fp= fdopen(fd,"r");
    }
    else
    {
	lock_exclusive(fd,item_file);
	fp= fdopen(fd,"r+");
    }

    /* Seek to the response */
    if (seek_resp(fp,resp) != resp)
    {
    	fclose(fp);
    	return(NULL);
    }

    /* Save the offset */
    off= ftell(fp);

    /* Read rest of ,R line and save rcode */
    if (fgets(bf,BFSZ,fp) == NULL)
    	die("EOF in response header");
    if (rcode != NULL) strncpy(rcode, bf, 4);

    /* Read ,U line and save uid and id */
    if (fgets(bf,BFSZ,fp) == NULL || bf[0] != ',' || bf[1] != 'U' ||
	(a= strchr(bf+2,',')) == NULL)
	die("bad response format");
    *uid= atoi(bf+2);
    if (id != NULL)
    {
	*firstin(a+1,"\n")= '\0';
	strncpy(id,a+1,MAX_LOGIN_LEN);
    }

    /* Seek back to just after ,R */
    fseek(fp, off, 0);
    return(fp);
}


/* HIDE_RESP - Mark a response hidden or unhidden.  Return 2 if no such item
 * or response, 1 if the response was already set that way or was scribbled,
 * or 0 on success.
 */

int hide_resp(int item, int resp, int hide)
{
FILE *fp;
char rcode[4];
uid_t ruid;
char rid[MAX_LOGIN_LEN];

    if (idict(VAR_UID) < 0 || sdict(VAR_ID)[0] == '\0')
	die("anonymous users cannot hide responses");

    if ((fp= open_resp(item,resp,1,rcode,rid,&ruid)) == NULL)
	return 2;

    if (ruid != idict(VAR_UID) || strncmp(rid,sdict(VAR_ID),MAX_LOGIN_LEN))
    {
	if (!bdict(VAR_AMFW))
	    die("attempt to hide a response not your own");
	  
	if (!bdict(VAR_AMADM))
	{
	    struct stat st;

	    if (!bdict(VAR_FW_HIDE))
		die("response hiding by fairwitnesses not enabled.");

	    if (fstat(fileno(fp),&st) || st.st_nlink > 1)
		die("fairwitness may not hide responses in linked items");
	}
    }
    else if (!bdict(VAR_AUTHOR_HIDE))
	die("response hiding by authors not enabled.");

    if (rcode[3] == '3' || (hide && rcode[3] != '0') ||
			   (!hide && rcode[3] != '1'))
    {
	fclose(fp);
	unlock(item_file);
	return(1);
    }

    rcode[3]= hide ? '1' : '0';

    fwrite(rcode, 1, 4, fp);

    fclose(fp);
    unlock(item_file);
    return(0);
}


/* ERASE_RESP - Erase a response.  Return 2 if there was no such item or
 * response, 1 if the response was already erased,  or 0 on success.
 */

int erase_resp(int item, int resp)
{
FILE *fp, *lfp;
char rcode[4];
uid_t ruid;
int i,j;
char bf[BFSZ+1];
long textoff;
size_t textlen;
time_t clock;
char *scrib[3];
#ifdef HIDE_HTML
int hidden_html;
#endif

    if (idict(VAR_UID) < 0 || sdict(VAR_ID)[0] == '\0')
	die("anonymous users cannot erase responses");

    if ((fp= open_resp(item,resp,1,rcode,bf,&ruid)) == NULL)
	return(2);

    if (ruid != idict(VAR_UID) || strncmp(bf,sdict(VAR_ID),MAX_LOGIN_LEN))
    {
	if (!bdict(VAR_AMFW))
	    die("attempt to erase a response not your own");

	if (!bdict(VAR_AMADM))
	{
	    struct stat st;

	    if (!bdict(VAR_FW_ERASE))
		die("response erasing by fairwitnesses not enabled.");

	    if (fstat(fileno(fp),&st) || st.st_nlink > 1)
		die("fairwitness may not censor linked items.");
	}
    }
    else if (!bdict(VAR_AUTHOR_ERASE))
	die("response erasing by authors not enabled.");

    /* Already erased - go home */
    if (rcode[3] == '3')
    {
	fclose(fp);
	unlock(item_file);
	return(1);
    }

#ifdef CENSOR_LOG
    /* Start writing the logfile entry */
    lfp= fopen(CENSOR_LOG,"a");
    if (lfp)
    {
	clock= time(0L);
#ifdef HIDE_HTML
	fprintf(lfp,"response %s:%d:%d (%.8s) censored by (%s) on %s",
#else
	fprintf(lfp,"response %s:%d:%d (%.8s) censored by (%s) on %s,T\n",
#endif /* HIDE_HTML */
		sdict(VAR_CONF), item, resp, bf, sdict(VAR_ID),
		ctimez(&clock,NULL));
    }
#endif /* CENSOR_LOG */

    /* Detect responses with hidden HTML */
#ifdef HIDE_HTML
    hidden_html= (rcode[2] != '0');
#ifdef CENSOR_LOG
    if (!hidden_html && lfp)
    	fputs(",T\n",lfp);
#endif /* CENSOR_LOG */
#endif /* HIDE_HTML */

    /* Mark it erased and plain-text */
    rcode[2]= '0';
    rcode[3]= '3';
    fwrite(rcode, 1, 4, fp);

    fseek(fp,1L,1);	/* we need a seek between reading and writing */

    /* Find start of response */
    for (;;)
    {
	if (fgets(bf,BFSZ,fp) == NULL)
	    die("Premature EOF in item file");
#ifdef HIDE_HTML
	if (hidden_html && bf[0] == ',' && bf[1] == 'D')
	    break;
#endif
	if (bf[0] == ',' && bf[1] == 'T')
	    break;
    }
    textoff= ftell(fp);

    /* Find end of response */
    do {
	if (fgets(bf,BFSZ,fp) == NULL)
	    die("Premature EOF in response text");
#ifdef CENSOR_LOG
	if (lfp) fputs(bf,lfp);
#endif
    } while (bf[0] != ',' || (bf[1] != 'E' && bf[1] != 'R'));
    textlen= ftell(fp) - strlen(bf) - textoff;
#ifdef CENSOR_LOG
    fclose(lfp);
#endif

    /* Make scribble string */
    clock= time(0L);
    scrib[0]= sdict(VAR_ID);
    scrib[1]= ctimez(&clock,NULL); scrib[1][24]= '\0';
    scrib[2]= sdict(VAR_ALIAS);
    for (i= j= 0; i < 75; j= (j+1)%3)
    {
	strcpy(bf+i, scrib[j]);
	i+= strlen(scrib[j]);
	bf[i++]= ' ';
    }
    bf[75]='\n';

    /* Return to start of text */
    fseek(fp,textoff,0);

#ifdef HIDE_HTML
    if (hidden_html)
    {
    	fwrite(",T\n", 1, 3, fp);
	textlen-= 3;
    }
#endif

    /* Overwrite text */
    while (textlen > 76)
    {
	fwrite(bf, 1, 76, fp);
	textlen-= 76;
    }
    if (textlen > 0)
    {
	bf[textlen-1]='\n';
	fwrite(bf, 1, textlen, fp);
    }

    fclose(fp);
    unlock(item_file);
    return(0);
}


/* ITEM_MODE - Mark an item as being retired or frozen.  Return 0 on success,
 * 2 if there is no such item, 3 if you don't have the proper access, 1 if
 * the item already has the given modes.
 */

int item_mode(int item, int im_cmd)
{
struct stat st;
FILE *fp;
char rid[MAX_LOGIN_LEN];
uid_t ruid;
int rc;

    /* Refuse anonymous users */
    if (idict(VAR_UID) < 0 || sdict(VAR_ID)[0] == '\0')
	return(3);

    if ((fp= open_resp(item,0,0,NULL,rid,&ruid)) == NULL)
	return(2);

    /* Stat the item */
    if (fstat(fileno(fp),&st))
	die("stat failed");

    /* Check if user has permission to do this operation */
    if (bdict(VAR_AMFW))
    {
    	/* Fairwitnesses can do it if the item is not linked */
    	if (!bdict(VAR_AMADM) && st.st_nlink > 1)
	    { rc= 3; goto done; }
    }
    else
    {
	/* Item authors can do it if the configuration says the can */
	if (ruid != idict(VAR_UID) ||
	    strncmp(rid,sdict(VAR_ID),MAX_LOGIN_LEN) || 
	    ((im_cmd == IM_FREEZE || im_cmd == IM_THAW) &&
	     !bdict(VAR_AUTHOR_FREEZE)) ||
	    ((im_cmd == IM_RETIRE || im_cmd == IM_UNRETIRE) &&
	     !bdict(VAR_AUTHOR_RETIRE)))
	    { rc= 3; goto done; }
    }

    rc= 1;
    switch(im_cmd)
    {
    case IM_FREEZE:
	if (!(st.st_mode & 0200)) goto done;
#ifdef CONF_LOG
	conflog(sdict(VAR_CONFDIR),"%s froze item %d",sdict(VAR_ID),item);
#endif
	st.st_mode&= ~0200;
	(void) put_sum(item, -1, SF_FROZEN, 0, 0L, 0L);
	break;
    case IM_THAW:
	if (st.st_mode & 0200) goto done;
#ifdef CONF_LOG
	conflog(sdict(VAR_CONFDIR),"%s thawed item %d",sdict(VAR_ID),item);
#endif
	(void) put_sum(item, -1, 0, SF_FROZEN, 0L, 0L);
	st.st_mode|= 0200;
	break;
    case IM_RETIRE:
	if (st.st_mode & 0100) goto done;
#ifdef CONF_LOG
	conflog(sdict(VAR_CONFDIR),"%s retired item %d",sdict(VAR_ID),item);
#endif
	st.st_mode|= 0100;
	(void) put_sum(item, -1, SF_RETIRED, 0, 0L, 0L);
	break;
    case IM_UNRETIRE:
	if (!(st.st_mode & 0100)) goto done;
#ifdef CONF_LOG
	conflog(sdict(VAR_CONFDIR),"%s unretired item %d",sdict(VAR_ID),item);
#endif
	st.st_mode&= ~0100;
	(void) put_sum(item, -1, 0, SF_RETIRED, 0L, 0L);
	break;
    }
    rc= 0;

    fchmod(fileno(fp), st.st_mode);
	
done:
    fclose(fp);
    unlock(item_file);

    return(rc);
}


/* KILL_ITEM - Kill an item.  Returns 0 on success, 2 if there is no such
 * item, 3 if you don't have the proper access.
 */

int kill_item(int item)
{
FILE *fp;
char rid[MAX_LOGIN_LEN];
uid_t ruid;
int nresp,flags,rc;
time_t idate, ldate;

    /* Deny anonymous users */
    if (idict(VAR_UID) < 0 || sdict(VAR_ID)[0] == '\0')
	return(3);

    if ((fp= open_resp(item,0,0,NULL,rid,&ruid)) == NULL)
	return(2);

    /* Can kill if we are FW, or if we are author and there are no responses */
    if (!bdict(VAR_AMFW))
    {
	rc= 3;
	if (ruid != idict(VAR_UID) ||
	    strncmp(rid,sdict(VAR_ID),MAX_LOGIN_LEN) ||
	    !bdict(VAR_AUTHOR_KILL))
	    goto done;
	rc= 2;
	if (get_sum(item, &nresp, &flags, &idate, &ldate))
	    goto done;
	rc= 3;
	if (nresp > 1)
	    goto done;
    }
#ifdef CONF_LOG
    conflog(sdict(VAR_CONFDIR),"%s killed item %d",sdict(VAR_ID),item);
#endif
    rc= 0;
    /* There is some risk of a response slipping in after this check */

    /* Nuke the file */
    unlink(item_file);

    /* Scrub it from sum file */
    (void) put_sum(item, 0, 0, -1, 0L, 0L);

    /* Nuke the index file */
    remove_index(sdict(VAR_CONFDIR),item);

done:
    fclose(fp);
    unlock(item_file);

    return(rc);
}



/* LINK_ITEM:  Create a link from the given item in the given conference to
 * the current conference.  Sets this_item to the new item number in this
 * conference and returns an error code:
 *    0 success
 *    1 cannot access other conference - it is private or doesn't exist.
 *    2 no such item
 *    3 no access - you are not fairwitness
 */

int link_item(int that_item, char *that_conf, int *this_item)
{
char *that_dir;
char this_file[BFSZ], that_file[BFSZ];
int mode;

    /* User must be fairwitness in this conference */
    if (!bdict(VAR_AMFW))
	return(3);

    /* Locate source conference */
    if ((that_dir= cfpath(that_conf)) == NULL)
	return(1);

    /* Check that source conference is not private */
    if (!bdict(VAR_AMADM))
    {
	readconfig(that_dir, NULL, NULL, &mode, NULL);
	mode&= 7;
	if (mode >= 4 && mode <= 6)
	    return(1);
    }

    /* Construct name of source item */
    sprintf(that_file,"%s/_%d", that_dir, that_item);

    /* Make the Link */
    *this_item= last_no() + 1;
    for (;;)
    {
	sprintf(this_file,"%s/_%d",sdict(VAR_CONFDIR), *this_item);
	if (!link(that_file,this_file))
		break;

	if (errno != EEXIST) return(2);
	(*this_item)++;
    }

    link_index(that_dir, that_item, sdict(VAR_CONFDIR), *this_item);
    link_sum(that_dir, that_item, sdict(VAR_CONFDIR), *this_item);

#ifdef CONF_LOG
    conflog(sdict(VAR_CONFDIR),"%s linked item %d from %s %d",
    	sdict(VAR_ID),*this_item,that_conf,that_item);
    conflog(that_dir,"%s linked item %d to %s %d",
    	sdict(VAR_ID),that_item,sdict(VAR_CONF),*this_item);
#endif

    return(0);
}



/* CONFLOG - Given the conference directory name and some printf() style
 * arguments (there should NOT be a terminating newline on the format string),
 * write a log message to the conference log file.  We don't do any locking,
 * but we do do the write in one call to minimize the damage.
 */

#ifdef CONF_LOG
#if __STDC__ && HAVE_STDVARARGS
void conflog(char *confdir, char *fmt, ...)
#else
void conflog(confdir, fmt, va_alist)
    char *confdir;
    char *fmt;
    va_dcl
#endif
{
#if HAVE_STDVARARGS
va_list ap;
#endif
    char fn[BFSZ+1];
    char *bf;
    FILE *fp;
    time_t now= time(0L);

    VA_START(ap,fmt);

    /* Generate name of log file */
    strcpy(fn,confdir);
    strcat(fn,"/log");

    /* Prefix format string with date and append newline */
    if ((bf= malloc(strlen(fmt)+25)) == NULL)
    	die("malloc failed in conflog()");
    strcpy(bf,ctime(&now)+4);
    bf[20]= ' ';
    strcpy(bf+21,fmt);
    strcat(bf,"\n");

    /* Do the write */
    if ((fp= fopen(fn,"a")) == NULL)
    	die("Cannot open log file %s",fn);
    VA_PRINTF(fp,bf,ap);
    fclose(fp);
    VA_END(ap);
}
#endif CONF_LOG
