/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* USER INFORMATION
 *
 * This file contains routines to access information about the user.  Many
 * sources are possible.  The system may be configured to use either or both
 * of two methods to identify the user:
 *
 *     ID_GETUID
 *        Primary user identification is taken from the getuid() call.  This
 *	  uid is used to directly access the other databases, unless
 *	  UID_MAP_FILE is defined, in which case it is mapped to a user name
 *	  using this file.
 *     ID_REMOTE_USER
 *        Primary user identification is taken from the "REMOTE_USER" 
 *	  environment variable, if and only iff getuid() returns SERVER_UID.
 *
 * The ID_GETUID define can be used with lynxcgi.  ID_REMOTE_USER is used
 * with standard .htaccess-like server authentication schemes.  If both are
 * defined, getuid() is used as the identification unless it returns
 * SERVER_UID, in which case the "REMOTE_USER" environment variable is used.
 * If neither is defined, authentication fails for all users, so they all
 * appear anonymous.
 *
 * Once we have the primary identification, this is used to look up all other
 * user information.  This maybe done using several configurations.  Only one
 * of the following defines may be enabled:
 *
 *  UD_PASSWD
 *      Secondary User Data is taken from the /etc/passwd file.  If
 *	PASSWD_FILE is defined, it is a file of the same format as the
 *	/etc/passwd_file to be used instead.
 *
 *  UD_DBM
 *      Secondary User Data is taken from the DBM file DBM_NAM_FILE or
 *      DBM_UID_FILE, depending on whether we have a name or a uid as our
 *	primary key.  The dbm file is keyed on the primary id.  The content
 *	looks like:
 *   
 *       <passwd>:<otherid>:<gid>:<fullname>:<directory>:<ignored>
 *
 *	Where
 *		<passwd> = encrypted user password.
 *		<otherid> = uid or id -- whichever was not the key.
 *		<gid> = group-id
 *		<fullname> = user's full name.
 *		<directory> = user's home directory.
 *		<ignored> = skip this field for now.
 *
 *	This could be the same file used as Apache's AuthDBMUserFile
 *	Actually, it would be weird to use this with ID_GETUID.
 *
 *  UD_TAGFILE
 *      There is no secondary user information database.  This does not work
 *	unless the primary ID is a name.  The user's directory is constructed
 *	directly from the user name using USER_DIR and USER_DIR_LEVEL.  All
 *	other information is taken from the IDENT_FILE maintained there by
 *	backtalk.  This configuration should not be used with real unix
 *	home directories.
 */


#include "backtalk.h"

#include <pwd.h>

#include "sysdict.h"
#include "last_log.h"
#include "stack.h"
#include "user.h"
#include "udb_read.h"
#include "udb_edit.h"
#include "udb_check.h"


struct passwd *upw;		/* NULL means user is anonymous */
char *effective_user= NULL;	/* set by "selectuser" */
char *new_fname= NULL;		/* If fullname has changed, this is it */




/* GETUSER - Identify the user and save his database informtion into the
 * upw variable.  On failure, we set upw to NULL.  We return a error code:
 *   0 - success
 *   1 - failure - no authentication - should treat as anonymous user
 *   2 - failure - authentication expired - should re-authenticate
 * This must be called before any call to user_fname(), user_id(), user_dir(),
 * user_uid() or user_gid() and before calls to func_changename(),
 * func_changepass() or func_removeuser() which apply to the current user.
 */

int getuser()
{
char *nam;
uid_t realuid;
int rc= 1;
#ifdef ID_SESSION
char bf[MAX_LOGIN_LEN+1];
#endif

    realuid= getuid();
    upw= NULL;		/* Yes, this is needed! */

#if defined(ID_REMOTE_USER) || defined(ID_SESSION)
#ifndef INSECURE
    if (realuid == SERVER_UID || realuid == 0)
#endif
    {
#ifdef ID_REMOTE_USER
	if ((nam= getenv("REMOTE_USER")) == NULL)
#endif
	{
#ifdef ID_SESSION
	    if ((rc= session_auth(nam= bf)) != 0)
#endif
		return rc;
	}

	upw= getdbnam(nam);
	return (upw == NULL);
    }
#endif /* ID_REMOTE_USER || ID_SESSION */

#ifdef ID_GETUID
#ifdef UID_MAP_FILE
    if ((nam= mapuid(realuid)) != NULL)
	upw= getdbnam(nam);
#else
    upw= getdbuid(realuid);
#endif /* UID_MAP_FILE */
#endif /* ID_GETUID */
    return (upw == NULL);
}



/* USER_FNAME - Return full name of the user.  Returns "" if hte user is
 * anonymous.
 */

char *user_fname()
{
    return (new_fname == NULL ?
		(upw == NULL ?  "" : upw->pw_gecos) : new_fname);
}


/* USER_ID - Return login id of the user.  Returns "" if the user is anonymous.
 */

char *user_id()
{
    return (upw == NULL ? "" : upw->pw_name);
}


/* USER_DIR - Return home directory of the user.  Returns "" if the user is
 * anonymous.
 */

char *user_dir()
{
    return (upw == NULL ? "" : upw->pw_dir);
}


/* USER_UID - Return uid of the user.  Returns -1 if he is anonymous.
 */

uid_t user_uid()
{
    return (upw == NULL ? -1 : upw->pw_uid);
}


/* USER_ISADM - Return true if the user is a conference administrator.
 */

int user_isadm()
{
    return (upw != NULL && inlist(upw->pw_name,ADM_NAMES,","));
}

#ifdef REAL_GID
/* USER_GID - Return group id of the user.  Returns -1 if he is anonymous.
 */

gid_t user_gid()
{
    return (upw == NULL ? -1 : upw->pw_gid);
}
#endif


/* FUNC_USERINFO:  <username> userinfo <uid> <gid> <fullname> <dir> <status>
 *
 * Load info about an arbitrary user.  All fields are 0 or {} if the user
 * does not exist.
 */

void func_userinfo()
{
char *name= pop_string();
char *pass;
struct passwd *pw;

    pw= getdbnam(name);
    if (pw == NULL)
    {
	push_integer(0);
	push_integer(0);
	push_string("",TRUE);
	push_string("",TRUE);
	push_time(0);
	push_integer(0);
    }
    else
    {
	time_t laston= get_lastlog(pw->pw_uid);

	push_integer(pw->pw_uid);
	push_integer(pw->pw_gid);
	push_string(pw->pw_gecos,TRUE);
	push_string(pw->pw_dir,TRUE);
	push_time(laston);

	pass= getpassword(name);
	if (pass == NULL || pass[0] == '*')
	    push_integer(2);
	else
	    push_integer(pass[0] == '?');
    }
}


/* FUNC_FIRSTUSER: - firstuser <username>
 * This returns the name of the "first" user in the user database.  It can
 * be used, together with "nextuser" to walk through the list of users.  If
 * there are no users, an empty string is pushed.
 */

void func_firstuser()
{
char *login;

    if ((login= firstusr()) == NULL)
	push_string("",TRUE);
    else
	push_string(login,TRUE);
}


/* FUNC_NEXTUSER: - nextuser <username>
 * This returns the name of the "next" user in the user database.  It can
 * be used, together with "firstuser" to walk through the list of users.  If
 * there are no more users, an empty string is pushed.
 */

void func_nextuser()
{
char *login;

    if ((login= nextusr()) == NULL)
	push_string("",TRUE);
    else
	push_string(login,TRUE);
}


/* FUNC_SEEKUSER: <username> seekuser -
 * This sets the user database so that the next call to "nextuser" will return
 * the user *after* the named user.  It is somewhat slow on some types of
 * databases.
 */

void func_seekuser()
{
char *login= pop_string();
    seekusr(login);
    free(login);
}


/* FUNC_SELECTUSER: <loginname> selectuser -
 * Have changename, changepass, and write work on some other user's account
 * instead of your own.  Only the conference administrator can execute this.
 * If given a null string, it reverts to the real user.
 */

void func_selectuser()
{
    if (!bdict(VAR_AMADM))
	die("Only conference administrator may do a selectuser");
    
    if (effective_user != NULL)
	free(effective_user);
    
    effective_user= pop_string();
    if (effective_user[0] == '\0')
    {
	free(effective_user);
	effective_user= NULL;
    }
    else
    {
    	if (getdbnam(effective_user) == NULL)
	    die("User %s does not exist",effective_user);
    }
}


/* FUNC_CHANGENAME: <fullname> changename <err>
 * Change the user's full name.  Push an error message if name is bad, or a
 * null string on success.
 */

void func_changename()
{
char *newname;
char err[BFSZ];
struct passwd *pw;

#if NOPWEDIT
    die("Cannot change fullnames on real unix accounts");
#else
    if (effective_user == NULL)
    {
	pw= upw;
	if (pw == NULL)
	    die("Cannot change name of unauthenticated user");
    }
    else
	if ((pw= getdbnam(effective_user)) == NULL)
	    die("Selected user does not exist");

    newname= pop_string();
    if (fullname_bad(newname,err))
    {
	push_string(err, TRUE);
	return;
    }

    change_fname(pw,newname);

    if (effective_user == NULL)
    {
	if (new_fname != NULL) free(new_fname);
	new_fname= newname;
    }
    else
	free(newname);

    push_string("",TRUE);
#endif
}


/* FUNC_CHANGEPASS: <password> changepass <err>
 * Change the user's password.  Push an error message if password is no good,
 * or a null string if it is OK.  This validates the account as a side effect.
 */

void func_changepass()
{
char err[BFSZ];
char *newpw;
struct passwd *pw;

#if NOPWEDIT
    die("Cannot change passwords on real unix accounts");
#else
    if (effective_user == NULL)
    {
	pw= upw;
	if (pw == NULL)
	    die("Cannot change password of unauthenticated user");
    }
    else
	if ((pw= getdbnam(effective_user)) == NULL)
	    die("Selected user does not exist");

    newpw= pop_string();
    if (passwd_bad(newpw,pw->pw_name,err))
    {
	push_string(err, TRUE);
	return;
    }
    change_passwd(pw,mkpasswd(newpw,0));
    free(newpw);
    push_string("",TRUE);
#endif
}


/* FUNC_REMOVEUSER: - removeuser -
 * Remove the user's account.  Bombs on errors.
 */

void func_removeuser()
{
char err[BFSZ];
struct passwd *pw;
#if NOPWEDIT
    die("Cannot remove real unix accounts");
#else

    if (effective_user == NULL)
    {
	pw= upw;
	if (pw == NULL)
	    die("Cannot delete unauthenticated user");
	upw= NULL;		/* Invalidate user data */
    }
    else
	if ((pw= getdbnam(effective_user)) == NULL)
	    die("Selected user does not exist");

    kill_user(pw);
#endif
}


/* EFFECTIVE_DIR - Return the home directory of the effective user.
 */

char *effective_dir()
{
    if (effective_user == NULL)
	return sdict(VAR_HOMEDIR);
    else
    {
	/* We don't call gethomedir here because we already called getdbname
	 * when we did "selectuser" so this data is in cache and this is
	 * actually faster than gethomedir().
	 */
	struct passwd *pw= getdbnam(effective_user);
	return pw->pw_dir;
    }
}


/* FUNC_VALIDATE: <login> <flag> validate -
 *
 * Change the status of an account.  Validates if <flag> is true, invalidates
 * if <flag> is false.
 */

void func_validate()
{
int flag, i= 0;
char *id, *oldpw;
char newpw[BFSZ];
struct passwd *pw;

#if NOPWEDIT
    die("Cannot change validation status of real unix accounts");
#else
    if (!bdict(VAR_AMADM))
	die("user is not a conference adminstrator");

    flag= pop_integer();
    id= pop_string();

    if ((oldpw= getpassword(id)) == NULL || (pw= getdbnam(id)) == NULL)
	die("user %s does not exist",id);
    if (oldpw[0] == '*' || oldpw[0] == '?') oldpw++;
    
    if (!flag) newpw[i++]= '*';
    strcpy(newpw+i, oldpw);

    change_passwd(pw, newpw);
#endif
}
